try:
    from Trader import Trader
    from Convoy import Convoy
    from Saveable import Saveable
    from Commodity import Stock
    import Market
    from db import marketdb
except:
    from mural.Trader import Trader
    from mural.Convoy import Convoy
    from mural.Saveable import Saveable
    from mural.Commodity import Stock
    from mural.Market import Market
    from mural.db import marketdb
_business=marketdb.business

class Business(Trader,Saveable):
    def all():
        cursor=_business.find({})
        output=[]
        for b in cursor:
            output.append(b)
        return output
    
    def __init__(self,name,market,money=0):
        super(Business,self).__init__(name,market,money)
        market.addBusiness(self)
        self.convoys=[]
        
    def load(d,market=None):
        d=Saveable.load(d,_business)
        if d is None:
            print("This business does not exist! Check spelling; it is case sensitive!")
            return None
        if market is None:
            market=Market.load(d['market'],deep=False)
        print("Getting %s's stock info..."%d['name'])
        b=Business(d['name'],market,d['money'])
        b.stock={name: Stock.load(s) for name,s in d['stock'].items()}
        return b
    def save(self):
        super(Business,self).save(_business)

    def display(self,commodity=None):
        print("|===== {:^11.10} =====|".format(self.name))
        print("|  Money      ${:<8.2f} |".format(self.money))
        print("|-------  Stock  -------|")
        print("|                       |")
        stocks=self.stock.items()
        if commodity is not None:
            stocks=[(commodity,self.stock[commodity])]
        for name,stock in stocks:
            print("|****   {:^9.6}   ****|".format(name))
            print("| Quantity:    {:<8d} |".format(stock.quantity))
            if stock.buyPrice > 1000:
                print("| Buy Price:  ${:06.2e} |".format(stock.buyPrice))
            else:
                print("| Buy Price:  ${:<8.2f} |".format(stock.buyPrice))
            if stock.sellPrice > 1000:
                print("| Sell Price: ${:06.2e} |".format(stock.sellPrice))
            else:
                print("| Sell Price: ${:<8.2f} |".format(stock.sellPrice))
            print("|                       |")
        print("|=======================|\n")
        
    def data(self):
        data=super(Business,self).data()
        data['convoys']=len(self.convoys)
        data['market']=self.market._id
        return data
    
    def createConvoy(self,cost,money=0,quantity=0):
        export=self.market.getMainExport()
        convoy=Convoy(self,cost,money,quantity)
        self.money-=money+cost
        self.stock[export].quantity-=quantity
        self.convoys.append(convoy)
        return convoy
    




try:
    from Commodity import Commodity,Stock
except:
    from mural.Commodity import Commodity,Stock
    
class Trader:
    def __init__(self,name,market,money=0):
        super(Trader,self).__init__()
        self.name=name
        self.market=market
        self.money=money
        stock={c.name: c.getStock() for c in Commodity.all()}
        self.stock=stock

    def data(self):
        return {
            'name': self.name,
            'money': self.money,
            'stock': {stock.name: stock.data() for key,stock in self.stock.items()}
        }
    
    def setMarket(self,market):
        self.market=market
        
    def __repr__(self):
        return "%s (%s)"%(self.name,self.market)

    def setBuyPrice(self,commodity,price):
        if issubclass(type(commodity),Commodity):
            commodity=commodity.name
        self.stock['commodity'].buyPrice=price
        
    def setSellPrice(self,commodity,price):
        if issubclass(type(commodity),Commodity):
            commodity=commodity.name
        self.stock['commodity'].buyPrice=price

    def spend(self,money,supress=False):
        if self.money-money < 0:
            if not supress: print("Cannot transact! Not enough money.")
            return False
        else:
            self.money-=money
            return True

    #Assume a transaction from A to B
    def transact(A,B,commodity,units,buy=None,sell=None,supress=False):
        if units <= 0:
            print("Wrong parameters in transaction.")
            return None
        if not issubclass(Commodity,type(commodity)):
            raise Exception("Old transaction occured! Please update to support commodities.")
        if A.market is None or B.market is None or A.market.name != B.market.name:
            print("Wrong market; no sale.",A,B)
            return None

        stockA=A.stock[commodity.name]
        stockB=B.stock[commodity.name]

        if not supress:
            print("")
            print("|============= Transaction ============|")
            print("| Commodity {:>18.18s} pigment |".format(commodity.name))
            print("| {:20.20} {:>6d} pigments |".format(B.name+"'s Quantity",stockB.quantity))
            print("| Your Quantity {:>13d} pigments |".format(stockA.quantity))
        
        buyPrice=(stockA.buyPrice-stockB.sellPrice)/2 + stockB.sellPrice
        sellPrice=(stockB.buyPrice-stockA.sellPrice)/2 + stockA.sellPrice

        #Check if they are willing and if they have the quantity to sell
        #TODO: TEST: If I want 5 but they only have 4, buy the 4.
        _canBuy=stockA.buyPrice >= stockB.sellPrice
        canBuy=_canBuy and stockB.quantity>0
        _canSell=stockB.buyPrice >= stockA.sellPrice
        canSell=_canSell and stockA.quantity>0
        
        buyEnabled= buy is None or buy is True
        sellEnabled= sell is None or sell is True

        isOnlySell= sell is True and buy is None
        isOnlyBuy = buy is True and sell is None

        purchase=canBuy and buyEnabled and not isOnlySell
        sale=canSell and sellEnabled and not isOnlyBuy
        yesno=lambda a: 'Yes' if a else 'No'
        if not supress:
            purchase_text="Failed"
            sale_text="Failed"
            if purchase: purchase_text="Succeded"
            if sale: sale_text="Succeded"
            print("|                                      |")
            if not isOnlySell:
                print("| - - - - - - - Purchase - - - - - - - |")
                print("| Your            Buy Price  ${:<8.2f} |".format(stockA.buyPrice))
                print("| {:9.9s}      Sell Price  ${:<8.2f} |".format(B.name+"'s",stockB.sellPrice))
                print("| Is Buy Price >= Sell Price?      {:>3s} |".format(yesno(_canBuy)))
                print("| Does {:7.7s} have stock to sell? {:>3s} |".format(B.name,yesno(stockB.quantity>0)))
                print("| Purchase {:>27s} |".format(purchase_text))
                print("|                                      |")
            if not isOnlyBuy:
                print("| - - - - - - - - Sale - - - - - - - - |")
                print("| {:10.10s}       Buy Price ${:<8.2f} |".format(B.name+"'s",stockB.buyPrice))
                print("| Your            Sell Price ${:<8.2f} |".format(stockA.sellPrice))
                print("| Is Buy Price >= Sell Price?      {:>3s} |".format(yesno(_canSell)))
                print("| Do you have stock to sell?       {:>3s} |".format(yesno(stockA.quantity>0)))
                print("| Sale {:>31s} |".format(sale_text))
                print("|                                      |")
            
        if purchase:
            if stockB.quantity-units<0:
                return Trader.transact(A,B,commodity,stockB.quantity,buy,sell,supress=True)
            if A.spend(buyPrice*units,True):
                stockA.quantity+=units
                stockB.quantity-=units
                B.money+=buyPrice*units
                print("| - - - - - - - Conclusion - - - - - - |")
                print("| {:10.10s} recieved ${:<8.2f}        |".format(B.name,buyPrice*units))
                print("| You recieved {:5d} {:<17.17s} |".format(units,commodity.name+' pigments'))
                print("|======================================|")
                return True
            else:
                print("| - - - - - - - Conclusion - - - - - - |")
                print("| Sale failed, because you             |") 
                print("| do   not have enough money.          |")
                print("| You have ${:<8.2f}                   |".format(B.money))
                print("|======================================|")
        elif sale:
            if stockA.quantity-units<0:
                return Trader.transact(A,B,commodity,stockA.quantity,buy,sell,supress=True)
            if B.spend(sellPrice*units,True):
                stockA.quantity-=units
                A.money+=sellPrice*units
                stockB.quantity+=units
                print("| - - - - - - - Conclusion - - - - - - |")
                print("| {:8.8s} recieved {:3d} {:5.5s} pigments |".format(B.name,units,commodity.name))
                print("| You recieved ${:<22.2f} |".format(sellPrice*units))
                print("|======================================|")
                return True
            else:
                print("| - - - - - - - Conclusion - - - - - - |")
                print("| Sale failed, because {:<10.10s}      |".format(B.name)) 
                print("| does not have enough money.          |")
                print("| {:<10.10s} has ${:<8.2f}             |".format(B.name,B.money))
                print("|======================================|")
        else:
            #Failed transaction.
            print("| - - - - - - - Conclusion - - - - - - |")
            print("| No transaction occured.              |")

            if not _canBuy and not isOnlySell:
                print("| Consider raising your buy price.     |")
            if not _canSell and not isOnlyBuy:
                print("| Consider lowering your sell price.   |")
            if not _canBuy or not _canSell:
                print("| Do this with the update command.     |")

            if stockA.quantity == 0 and not isOnlyBuy:
                print("| Make sure that you have pigments to  |")
                print("| sell to your customer.               |")
            if stockB.quantity == 0 and not isOnlySell:
                print("| Your vendor does not have pigments   |")
                print("| To sell.                             |")
            print("|======================================|")
            return False
    
    #self is buying from business
    def buy(self,business,commodity,units):
        return Trader.transact(self,business,commodity,units,buy=True)

    #business is buying from self
    def sell(self,business,commodity,units):
        return Trader.transact(self,business,commodity,units,sell=True)

if __name__=='__main__':
    from Market import Market
    red=Commodity('Red')
    mall=Market('test',exports=[red])
    joe=Trader('joe',mall,money=10)
    joe.stock['Red']=Stock(red,1,10,10)
    ken=Trader('ken',mall,money=10)
    ken.stock['Red']=Stock(red,1,1,1)
    Trader.sell(joe,ken,red,20)

    



import bson
class Saveable:
    def __init__(self):
        self._id=None
    def load(d,db):
        obj=None
        if issubclass(str,type(d)):
            d=d.title()
            obj=db.find_one({'name': d})
        elif issubclass(bson.objectid.ObjectId,type(d)):
            obj=db.find_one({'_id': d})
        return obj
    def getId(name,db):
        response=db.find_one({'name': name})
        if response is not None:
            return response['_id']
        else: return None
    def data():
        return {'name': 'Aurther'}
    def save(self,db,v=False):
        data=self.data()
        db.update({'name': self.name},{'$set': data},upsert=True)
        self._id=db.find_one({'name': self.name})['_id']
        if v: print("Successful save of",self)



#Script used for building the environment.

try:
    from Market import Market
    from Commodity import Commodity
    from Business import Business #cluster0-7aeh1.azure.mongodb.net
except:
    from mural.Market import Market
    from mural.Commodity import Commodity
    from mural.Business import Business #cluster0-7aeh1.azure.mongodb.net

from pymongo import MongoClient
client=MongoClient('mongodb+srv://sailor:O9d4*eBJkizk@cluster0-7aeh1.azure.mongodb.net/test?retryWrites=true&w=majority')
marketdb=client.market_test
business=marketdb.business
market=marketdb.market
canvas=marketdb.canvas
transactions=marketdb.canvas_transactions

def buildCommodities():
    from Commodity import Commodity
    red=Commodity('red')
    blue=Commodity('blue')
    green=Commodity('green')
    red.save()
    blue.save()
    green.save()

def buildMarkets():
    red=Commodity('Red')
    green=Commodity('Green')
    blue=Commodity('Blue')
    for c in [red,green,blue]: c.save()
    
    sc=Market("Red", (5,0), exports=[red])
    sf=Market('Green',(-2,7),exports=[green])
    rwc=Market('Blue',(10,-8),exports=[blue])

    nike=Business("Nike",sc,100)
    adidas=Business("Adidas",sc,77)
    asic=Business('Asic',sf,200)
    reebok=Business('Reebok',sf,200)
    sketchers=Business('Sketchers',rwc,200)
    puma=Business('Puma',rwc,200)
    vans=Business('Vans',sf,200)
    businesses=[nike,adidas,asic,reebok,sketchers,puma,vans]
    for b in businesses:
        b.save()
                
    nike.stock['Red'].quantity+=5
    nike.stock['Blue'].quantity+=5
    adidas.stock['Red'].quantity+=5
    adidas.stock['Green'].quantity+=5
    sc.save()
    sf.save()
    rwc.save()

buildMarkets()

#test=Market.load("Santa Cruz")
#print(test,test.businesses)



try:
    from db import marketdb
except:
    from .db import marketdb
def setPoint(db,x,y,r,g,b,index=0):
    color={
        'x': x,
        'y': y,
        'r': r,
        'g': g,
        'b': b,
        'index': index,
    }
    return db.update_many({'x': x, 'y': y}, {'$set': color},upsert=True)

def buildCanvas(db,hue=10,size=50):
    _index=0
    for i in range(size):
        for j in range(size):
            _index+=1
            print(i,j,setPoint(db,i,j,hue,hue,hue,_index))
canvas=marketdb.canvas
buildCanvas(canvas,70,25)





from numpy.random import normal
from threading import Timer
import time,random,math
try:
    from Trader import Trader
    from Commodity import Stock
except:
    from mural.Trader import Trader
    from mural.Commodity import Stock

class Convoy(Trader):
    def __init__(self,owner,cost,money=0,quantity=0):
        name="%s's Convoy"%(owner.name)
        self.origin=owner.market
        self.owner=owner
        self.commodity=self.origin.getMainExport()
        
        buy=self.owner.stock[self.commodity].buyPrice
        sell=self.owner.stock[self.commodity].sellPrice
        
        stock=Stock(self.commodity,quantity,buy,sell)
        super(Convoy,self).__init__(name,owner.market,money)
        self.stock[stock.name]=stock
        self.cost=cost
        self.quantity=quantity
        self.speed=self.getSpeed(cost)
        self.destination=None
        
        self.businesses=[]
        print("Convoy created, ",self)
        
    def getSpeed(self,money):
        slowest=.5
        fastest=2
        scale=money/300
        if scale > 1: scale=1
        loc=slowest + (fastest-slowest) * scale
        spread=slowest*(1-scale)
        speed = normal(loc,spread)
        if speed > fastest: return fastest
        if speed < slowest: return slowest
        return speed

    def getTransitTime(self,market):
        return self.market.getDistance(market)/self.speed
    
    def gotoMarket(self,destination,businesses):
        if len(businesses) < 1:
            print("Must have specify which businesses to transact with.")
            return
        self.businesses=businesses
        self.setMarket(destination,self.arrivedRemote)


    #businesses is a tuple array (busisness,quantity)
    def setMarket(self,destination,finished):
        transitTime=self.getTransitTime(destination)
        print("%s going to %s will take %.02f sec."%(self,destination,transitTime))
        self.destination=destination
        self.market=None

        duration=transitTime
        steps=math.ceil(1.5*duration)
        char=str(random.choice(['Þ','.','*','-','Ø','o','#','!','^','|']))
        direction=random.choice(['>','<','^'])
        allSteps=range(steps)
        if random.choice([True,False]):
            allSteps=reversed(allSteps)
        for i in allSteps:
            print(("{:%s%ss}"%(direction,steps)).format(char*(i+1)))
            time.sleep(duration/steps)
        finished()

    def goHome(self):
        self.setMarket(self.origin,self.arrivedHome)
        
    def arrivedRemote(self):
        print(self,"arrived.")
        self.market=self.destination
        self.destination=None
        export=self.market.getMainExport()
        for business,quantity in self.businesses:
            Trader.transact(self,business,business.stock[export].commodity,quantity)
            time.sleep(2)
        self.goHome()
        
    def arrivedHome(self):
        if self.destination is not self.origin:
            raise Exception("Error wrong market")
        self.market=self.destination
        self.destination=None
        self.owner.money += self.money
        self.money=0
        for name,stock in self.stock.items():
            self.owner.stock[name].quantity += stock.quantity
            stock.quantity=0
        print(self.owner.convoys)
        del self.owner.convoys[self.owner.convoys.index(self)]
        print(self.owner.convoys)

